#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""

"""
from __future__ import print_function, division, absolute_import
import six
from six.moves.urllib.parse import urljoin
import re
import os
import pickle
import logging
import json as jsonlib
import requests
from requests.auth import HTTPBasicAuth

from requests import Session
import websocket

logger = logging.getLogger(__name__)


OPCODE_DATA = (websocket.ABNF.OPCODE_TEXT, websocket.ABNF.OPCODE_BINARY)


class KaeAPIError(Exception):
    def __init__(self, http_code, msg):
        self.http_code = http_code
        self.msg = msg

    def __str__(self):
        return self.msg


def recv(ws):
    try:
        frame = ws.recv_frame()
    except websocket.WebSocketException:
        return websocket.ABNF.OPCODE_CLOSE, None
    if not frame:
        raise websocket.WebSocketException("Not a valid frame %s" % frame)
    elif frame.opcode in OPCODE_DATA:
        return frame.opcode, frame.data
    elif frame.opcode == websocket.ABNF.OPCODE_CLOSE:
        ws.send_close()
        return frame.opcode, None
    elif frame.opcode == websocket.ABNF.OPCODE_PING:
        ws.pong(frame.data)
        return frame.opcode, frame.data

    return frame.opcode, frame.data


def recv_ws(ws):
    while True:
        opcode, data = recv(ws)
        msg = None
        if six.PY3 and opcode == websocket.ABNF.OPCODE_TEXT and isinstance(data, bytes):
            data = str(data, "utf-8")
        if opcode in OPCODE_DATA:
            msg = data

        if msg is not None:
            yield msg

        if opcode == websocket.ABNF.OPCODE_CLOSE:
            print('closed')
            break


class KaeAPI:
    def __init__(self, host, version='v1', timeout=None,
                 access_token=None, cluster='default'):
        self.host = host
        self.version = version
        self.timeout = timeout
        self.cluster = cluster

        self.host = host
        self.base = '%s/api/%s/' % (self.host, version)
        self.session = Session()

        if access_token:
            self.session.headers.update({
                'Authorization': 'Bearer {}'.format(access_token),
            })

    def set_access_token(self, access_token):
        self.session.headers.update({
            'Authorization': 'Bearer {}'.format(access_token),
        })

    def set_base_auth(self, user, password):
        auth = HTTPBasicAuth(user, password)
        auth(self.session)

    def set_real_user(self, real_user):
        self.session.headers.update({'X-Real-User': real_user})

    def set_cluster(self, cluster):
        self.cluster = cluster

    def request(self, path, method='GET', params=None, data=None, json=None, files=None, **kwargs):
        """Wrap around requests.request method"""
        url = urljoin(self.base, path)
        params = params or {}
        resp = self.session.request(url=url,
                                    method=method,
                                    params=params,
                                    data=data,
                                    json=json,
                                    timeout=self.timeout,
                                    files=files,
                                    **kwargs)
        code = resp.status_code
        if code != 200:
            raise KaeAPIError(code, resp.text)
        try:
            response = resp.json()
        except ValueError:
            # when console don't return json, it means a bug
            raise KaeAPIError(500, 'BUG: Console did not return json, body {}'.format(resp.text))
        return response

    def request_ws(self, path, params=None, data=None, json=None, ignore_decode_err=False):
        url = urljoin(self.base, path)
        url = re.sub(r'^http', 'ws', url)
        headers = {
            'Authorization': self.session.headers['Authorization'],
        }
        if 'X-Real-User' in self.session.headers:
            headers['X-Real-User'] = self.session.headers['X-Real-User']

        options = {
            'header': headers,
        }

        ws = websocket.create_connection(url, **options)
        ws.send(jsonlib.dumps(json))
        for msg in recv_ws(ws):
            try:
                data = jsonlib.loads(msg)
                yield data
            except jsonlib.JSONDecodeError:
                if ignore_decode_err:
                    # print("decode json error, ignore it")
                    continue
                else:
                    raise KaeAPIError(500, "json decode error {}".format(msg))
            except (ValueError, TypeError):
                raise KaeAPIError(500, msg)

    def list_app(self):
        return self.request('app/')

    def get_app(self, appname):
        return self.request('app/%s' % appname)

    def delete_app(self, appname):
        return self.request('app/%s' % appname, method='DELETE')

    def get_app_users(self, appname):
        return self.request('app/%s/users' % appname)

    def grant_user(self, appname, username=None, email=None):
        params = {}
        if username is not None:
            params['username'] = username
        elif email is not None:
            params['email'] = email
        else:
            raise ValueError("you need pass username or email")
        return self.request('app/%s/users' % appname, method='PUT', json=params)

    def revoke_user(self, appname, username=None, email=None):
        params = {}
        if username is not None:
            params['username'] = username
        elif email is not None:
            params['email'] = email
        else:
            raise ValueError("you need pass username or email")
        return self.request('app/%s/users' % appname, method='DELETE', json=params)

    def get_app_pods(self, appname, canary=False, watch=False):
        params = {
            'cluster': self.cluster,
            'canary': canary,
        }
        if watch is False:
            return self.request('app/%s/pods' % appname, params=params)
        else:
            return self.request_ws('ws/app/%s/pods/events' % appname, json=params)

    def watch_app_pods(self, appname, canary=False):
        payload = {
            'cluster': self.cluster,
            'canary': canary,
        }
        return self.request_ws('ws/app/%s/pods/events' % appname, json=payload)

    def get_app_releases(self, appname):
        return self.request('app/%s/releases' % appname)

    def get_app_deployment(self, appname, canary=False):
        params = {
            'cluster': self.cluster,
            'canary': canary,
        }
        return self.request('app/%s/deployment' % appname, params=params)

    def get_release(self, appname, tag):
        return self.request('app/%s/version/%s' % (appname, tag))

    def get_secret(self, appname):
        params = {
            'cluster': self.cluster,
        }
        return self.request('app/%s/secret' % appname, params=params)

    def set_secret(self, appname, data, replace=False):
        payload = {
            'cluster': self.cluster,
            'data': data,
            'replace': replace,
        }
        return self.request('app/%s/secret' % appname, method="POST", json=payload)

    def get_config(self, appname):
        params = {
            'cluster': self.cluster,
        }
        return self.request('app/%s/configmap' % appname, params=params)

    def set_config(self, appname, data, replace=False):
        payload = {
            'cluster': self.cluster,
            'data': data,
            'replace': replace,
        }
        return self.request('app/%s/configmap' % appname, method="POST", json=payload)

    def list_app_yaml(self, appname):
        return self.request('app/%s/yaml' % appname)

    def create_app_yaml(self, appname, name, specs_text, comment=None):
        params = {
            'name': name,
            'specs_text': specs_text,
        }
        if comment is not None:
            params['comment'] = comment
        return self.request('app/%s/yaml' % appname, method='POST', json=params)

    def update_app_yaml(self, appname, name, new_name, specs_text, comment=None):
        params = {
            'name': new_name,
            'specs_text': specs_text,
        }
        if comment is not None:
            params['comment'] = comment
        return self.request('app/%s/name/%s/yaml' % (appname, name), method='POST', json=params)

    def delete_app_yaml(self, appname, name):
        return self.request('app/%s/name/%s/yaml' % (appname, name), method='DELETE')

    def register_release(self, appname, tag, git, specs_text, branch=None, force=False):
        payload = {
            'appname': appname,
            'tag': tag,
            'git': git,
            'specs_text': specs_text,
            'branch': branch,
            'force': force,
        }
        return self.request('app/register', method='POST', json=payload)

    def rollback(self, appname, revision=0, deploy_id=None):
        payload = {
            'cluster': self.cluster,
            'revision': revision,
        }
        if deploy_id is not None:
            payload['deploy_id'] = deploy_id
        return self.request('app/%s/rollback' % appname, method='PUT', json=payload)

    def renew(self, appname):
        payload = {
            'cluster': self.cluster,
        }
        return self.request('app/%s/renew' % appname, method='PUT', json=payload)

    def build_app(self, appname, tag, block=False, ignore_decode_err=True):
        payload = {
            'tag': tag,
            'block': block,
        }
        return self.request_ws('ws/app/%s/build' % appname, json=payload,
                               ignore_decode_err=ignore_decode_err)

    def deploy_app(self, appname, tag, cpus=None, memories=None, replicas=None, app_yaml_name=None, use_newest_config=False):
        """deploy app.
        appname:
        tag: 要部署的版本号, git tag的值.
        cpus: 需要的cpu个数, 例如1, 或者1.5, 如果是public的部署, 传0.
        memories: 最小4MB.
        replicas: app的副本数量
        app_yaml_name: AppYaml template name
        """
        payload = {
            'cluster': self.cluster,
            'tag': tag,
            'cpus': cpus,
            'memories': memories,
            'replicas': replicas,
            'app_yaml_name': app_yaml_name,
            'use_newest_config': use_newest_config,
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        return self.request('app/%s/deploy' % appname, method='PUT', json=payload)

    def undeploy_app(self, appname, cluster=None):
        """undeploy app.
        appname: App name
        """
        if cluster is None:
            cluster = self.cluster
        payload = {
            'cluster': cluster,
        }
        return self.request('app/%s/undeploy' % appname, method='DELETE', json=payload)

    def deploy_app_canary(self, appname, tag, cpus=None, memories=None, replicas=None, app_yaml_name=None, use_newest_config=False):
        """deploy canary version of specified app
        appname:
        tag: 要部署的版本号, git tag的值.
        cpus: 需要的cpu个数, 例如1, 或者1.5, 如果是public的部署, 传0.
        memories: 最小4MB.
        replicas: app的副本数量
        app_yaml_name: AppYaml template name
        """
        payload = {
            'cluster': self.cluster,
            'tag': tag,
            'cpus': cpus,
            'memories': memories,
            'replicas': replicas,
            'app_yaml_name': app_yaml_name,
            'use_newest_config': use_newest_config,
        }

        payload = {k: v for k, v in payload.items() if v is not None}
        return self.request('app/%s/canary/deploy' % appname, method='PUT', json=payload)

    def undeploy_app_canary(self, appname, cluster=None):
        """undeploy canary version of specified app
        appname:
        """
        if cluster is None:
            cluster = self.cluster
        payload = {
            'cluster': cluster,
        }

        payload = {k: v for k, v in payload.items() if v is not None}
        return self.request('app/%s/canary/undeploy' % appname, method='DELETE', json=payload)

    def delete_app_canary(self, appname):
        payload = {
            'cluster': self.cluster,
        }
        return self.request('app/%s/canary' % appname, method='DELETE', json=payload)

    def scale_app(self, appname, replicas):
        """deploy app.
        replicas: app的副本数量
        """
        payload = {
            'cluster': self.cluster,
            'replicas': replicas,
        }

        return self.request('app/%s/scale' % appname, method='PUT', json=payload)

    def set_app_abtesting_rules(self, appname, rules):
        """deploy app.
        cpu: 需要的cpu个数, 例如1, 或者1.5, 如果是public的部署, 传0.
        memory: 最小4MB.
        replicas: app的副本数量
        """
        payload = {
            'cluster': self.cluster,
            'rules': rules,
        }
        return self.request('app/%s/abtesting' % appname, method='PUT', json=payload)

    def stop_container(self, appname, podname, container, namespace, cluster=None):
        if cluster is None:
            cluster = self.cluster
        payload = {
            'cluster': cluster,
            'podname': podname,
            'container': container,
            'namespace': namespace,
        }
        return self.request('app/%s/container/stop' % appname, method='POST', json=payload)

    def kill_build_task(self, appname):
        return self.request('app/%s/build/kill' % appname, method='DELETE')

    def create_job(self, specs_text=None, **kwargs):
        payload = kwargs
        if specs_text:
            payload = {
                'specs_text': specs_text,
            }
        return self.request('job', method='POST', json=payload)

    def list_job(self):
        return self.request('job', method='GET')

    def delete_job(self, jobname):
        return self.request('job/%s' % jobname, method='DELETE')

    def get_job_log(self, jobname, follow=False):
        if follow is False:
            return self.request('job/%s/log' % jobname)
        else:
            return self.request_ws('ws/job/%s/log/events' % jobname)

    def create_sparkapp(self, data=None, spec_text=None):
        payload = data
        if spec_text:
            payload = {
                'spec_text': spec_text,
            }
        return self.request('spark', method='POST', json=payload)

    def list_sparkapp(self):
        return self.request('spark', method='GET')

    def delete_sparkapp(self, appname):
        return self.request('spark/%s' % appname, method='DELETE')

    def restart_sparkapp(self, appname):
        return self.request('spark/%s/restart' % appname, method='PUT')

    def get_sparkapp_log(self, appname, follow=False):
        if follow is False:
            return self.request('spark/%s/log' % appname)
        else:
            return self.request_ws('ws/spark/%s/log/events' % appname)

    def upload(self, appname, filetype, files):
        data = {
            'fileType': filetype
        }

        return self.request('spark/%s/upload' % appname, data=data, files=files, method='POST')



# -*- coding: utf-8 -*-

import os
import re
from addict import Dict
from marshmallow import Schema, fields, validates_schema, ValidationError, post_load
from humanfriendly import parse_size, InvalidSize


_name_regex = re.compile(r'[a-z]([-a-z0-9]*[a-z0-9])?$')

class IntOrStrField(fields.Field):
    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, str) or isinstance(value, int):
            return value
        else:
            raise ValidationError('Field should be str or integer')


class StrictSchema(Schema):
    @validates_schema(pass_original=True)
    def check_unknown_fields(self, data, original_data):
        if original_data is None:
            raise ValidationError("the data passed the schema is null")
        unknown = set(original_data) - set(self.fields) - set(field.load_from for field in self.fields.values())
        if unknown:
            raise ValidationError('Unknown fields: {}, please check the docs'.format(unknown))

    class Meta:
        strict = True
        ordered = True


def validate_appname(name):
    if _name_regex.match(name) is None:
        raise ValidationError("appname is invalid")


def validate_jobname(name):
    if _name_regex.match(name) is None:
        raise ValidationError("jobname is invalid")


def validate_app_type(ss):
    if ss not in ("web", "worker"):
        raise ValidationError("app type should be `web`, `worker`")


def validate_spark_apptype(ss):
    if ss not in ("sparkapplication", "scheduledsparkapplication"):
        raise ValidationError("spark apptype should be `sparkapplication`, `scheduledsparkapplication`")


def validate_spark_type(ss):
    if ss not in ("Python", "Scala"):
        raise ValidationError("spark type should be `Python`, `Scala`")


def validate_spark_mode(mode):
    if mode not in ("client", "cluster"):
        raise ValidationError("spark mode should be `client`, `cluster`")


def validate_schedule(ss):
    #TODO
    pass


def validate_tag(tag):
    regex = re.compile(r'[\w][\w.-]{0,127}$')
    if regex.match(tag) is None:
        raise ValidationError("tag is invalid")


def validate_port(n):
    if not 0 < n <= 65535:
        raise ValidationError('Port must be 0-65,535')


def validate_protocol(ss):
    if ss not in ("TCP", "UDP"):
        raise ValidationError("protocol should be TCPor UDP")


def validate_env_list(l):
    for env in l:
        if len(env.split("=")) != 2:
            raise ValidationError("environment should conform to format: key=val")


def validate_image_pull_policy(ss):
    if ss not in ('Always', 'Never', 'IfNotPresent'):
        raise ValidationError("invalid imagePullPolicy value, only one of Always, Never, IfNotPresent is allowed")


def validate_abs_path(ss):
    if not os.path.isabs(ss):
        raise ValidationError("{} is not an absolute path".format(ss))


def validate_abs_path_list(lst):
    for l in lst:
        if l[0] != '/':
            raise ValidationError("{} is not a absolute path".format(l))


def validate_mountpoints(lst):
    hosts = set()
    for mp in lst:
        host = mp['host']
        if host in hosts:
            raise ValidationError('{} duplicate domain.'.format(host))
        hosts.add(host)


def validate_pod_volumes(lst):
    for vol in lst:
        if 'name' not in vol:
            raise ValidationError("need `name` field for volume")
        if 'persistentVolumeClaim' in vol:
            pvc = vol['persistentVolumeClaim']
            if not isinstance(pvc, dict):
                raise ValidationError("wrong PVC volume")
            if 'claimName' not in pvc:
                raise ValidationError("need `claimName` field for PVC volume")
        # notify use to use built-in Secret support
        if 'secret' in vol:
            raise ValidationError("please don't use Secret directly")
        # notify user to use built-in ConfigMap support
        if 'configMap' in vol:
            raise ValidationError("please don't use ConfigMap directly")
        if 'hostPath' in vol:
            host_path = vol['hostPath']
            if not isinstance(host_path, dict):
                raise ValidationError('wrong format for `hostPath` volumes')
            if 'path' not in host_path:
                raise ValidationError('need `path` field for hostPath volumes')
            inner_path = host_path['path']
            # 杜绝安全问题，避免瞎搞，比如将整个根目录挂载到容器(其实应该做的更严格，比如每个app一个目录)
            if not inner_path.startswith('/data/kae'):
                raise ValidationError('hostPath volume\'s path must be subdirectory of /data/kae')


def validate_build_name(name):
    if ':' in name:
        raise ValidationError("build name can't contain `:`.")


def validate_update_strategy_type(ss):
    if ss not in ("RollingUpdate", "Recreate"):
        raise ValidationError("strategy type must be RollingUpdate or Recreate")


def validate_percentage_or_int(ss):
    if ss.endswith('%'):
        ss = ss[:-1]
    try:
        if int(ss) < 0:
            raise ValidationError("must be positive number or zero.")
    except ValueError:
        raise ValidationError("invalid percentage or integer")


def validate_str_dict(dd):
    for k, v in dd.items():
        if not isinstance(k, str):
            raise ValidationError("key should be a string")
        if not isinstance(v, str):
            raise ValidationError("value should be a string")


def validate_cpu(d):
    for k, v in d.items():
        if k not in ('request', "limit"):
            raise ValidationError("cpu dict's key should be request or limit")
        try:
            if v[-1] == 'm':
                v = v[:-1]
            if float(v) < 0:
                raise ValidationError('CPU must >=0')
        except:
            raise ValidationError("invalid cpu value format")


def validate_memory(d):
    for k, v in d.items():
        if k not in ('request', "limit"):
            raise ValidationError("memory dict's key should be request or limit")
        try:
            if parse_size(v) <= 0:
                raise ValidationError("memory should bigger than zero")
        except InvalidSize:
            raise ValidationError("invalid memory value format")


def validate_docker_volumes(lst):
    for l in lst:
        if ':' not in l:
            raise ValidationError('wrong docker volumes')
        parts = l.split(':', 1)
        if (not os.path.isabs(parts[0])) or (not os.path.isabs(parts[1])):
            raise ValidationError('host path and container path must be absolute path')


class Mountpoint(StrictSchema):
    host = fields.Str(required=True)
    path = fields.Str(missing="/")
    paths = fields.List(fields.Str(), missing=['/'])
    tlsSecret = fields.Str()


class ContainerPort(StrictSchema):
    containerPort = fields.Int(validate=validate_port)
    hostIP = fields.Str()
    hostPort = fields.Int(validate=validate_port)
    name = fields.Str()
    protocol = fields.Str(validate=validate_protocol)


class BuildSchema(StrictSchema):
    name = fields.Str(validate=validate_build_name)
    tag = fields.Str(validate=validate_tag)
    dockerfile = fields.Str()
    target = fields.Str()
    args = fields.Dict()


class RollingUpdate(StrictSchema):
    maxSurge = fields.Str(missing="25%", validate=validate_percentage_or_int)
    maxUnavailable = fields.Str(missing="25%", validate=validate_percentage_or_int)


class UpdateStrategy(StrictSchema):
    type = fields.Str(missing="RollingUpdate", validate=validate_update_strategy_type)
    # only valid when type is RollingUpdate
    rollingUpdate = fields.Nested(RollingUpdate)


class ConfigMapSchema(StrictSchema):
    dir = fields.Str(required=True)
    key = fields.Str(required=True)
    filename = fields.Str()

    @post_load
    def add_defaults(self, data):
        if 'filename' not in data:
            data['filename'] = data['key']
        if not os.path.isabs(data['dir']):
            raise ValidationError("{} is not a absolute path".format(data['dir']))
        return data


class SecretSchema(StrictSchema):
    envNameList = fields.List(fields.Str(), required=True)
    keyList = fields.List(fields.Str())

    @post_load
    def add_defaults(self, data):
        if 'keyList' not in data:
            data['keyList'] = data['envNameList']
        if len(data['keyList']) != len(data['envNameList']):
            raise ValidationError("the length of envNameList must equal to keyList")
        return data


class VolumeMountSchema(StrictSchema):
    name = fields.Str(required=True)
    mountPath = fields.Str(required=True)
    readOnly = fields.Bool(missing=False)
    subPath = fields.Str(missing="")


build_schema = BuildSchema()


class ContainerSpec(StrictSchema):
    name = fields.Str()
    image = fields.Str()
    imagePullPolicy = fields.Str(validate=validate_image_pull_policy)
    args = fields.List(fields.Str())
    command = fields.List(fields.Str())
    env = fields.List(fields.Str(), validate=validate_env_list)
    tty = fields.Bool()
    workingDir = fields.Str(validate=validate_abs_path)
    livenessProbe = fields.Dict()
    readinessProbe = fields.Dict()
    ports = fields.List(fields.Nested(ContainerPort))

    cpu = fields.Dict(validate=validate_cpu)
    memory = fields.Dict(validate=validate_memory)
    gpu = fields.Int()

    configs = fields.List(fields.Nested(ConfigMapSchema), missing=[])
    secrets = fields.Nested(SecretSchema)
    volumeMounts = fields.List(fields.Nested(VolumeMountSchema), missing=[])
    useDFS = fields.Bool(missing=False)

    @post_load
    def add_defaults(self, data):
        if "cpu" not in data:
            cpu = {"request": "100m", "limit": "200m"}
            data['cpu'] = cpu
        if "memory" not in data:
            memory = {"request": "64Mi", "limit": "128Mi"}
            data["memory"] = memory
        return data


class ServicePort(StrictSchema):
    name = fields.Str()
    port = fields.Int(required=True, validate=validate_port)
    targetPort = IntOrStrField()
    protocol = fields.Str(validate=validate_protocol, missing="TCP")


class HostAliases(StrictSchema):
    ip = fields.Str(required=True)
    hostnames = fields.List(fields.Str(), required=True)


class HPAMetric(StrictSchema):
    name = fields.Str(required=True)
    averageUtilization = fields.Int()
    averageValue = fields.Str()
    value = fields.Str()

    @validates_schema(pass_original=True)
    def further_check(self, data, original_data):
        name = data.get("name")
        if name.lower() not in ("cpu", "memory"):
            raise ValidationError("name msut be cpu or mempry")
        averageUtilization = data.get("averageUtilization")
        averageValue = data.get("averageValue")
        value = data.get("value")

        none_field_cnt = [averageUtilization, averageValue, value].count(None)
        if none_field_cnt != 2:
            raise ValidationError('you must specify one and only one field of averageUtilization, averageValue and value')


class HPA(StrictSchema):
    minReplicas = fields.Int(missing=1)
    maxReplicas = fields.Int(required=True)
    metrics = fields.List(fields.Nested(HPAMetric))

    @validates_schema(pass_original=True)
    def further_check(self, data, original_data):
        minReplicas = data.get("minReplicas")
        maxReplicas = data.get("maxReplicas")
        if maxReplicas < minReplicas:
            raise ValidationError("maxReplicas must not be less than minReplicas")
        metrics = data.get("metrics")
        if metrics is None or len(metrics) == 0:
            raise ValidationError("at least one metric is needed for HPA")


class ServiceSchema(StrictSchema):
    user = fields.Str(missing="root")
    registry = fields.Str()
    labels = fields.List(fields.Str())
    ingressAnnotations = fields.Dict(validate=validate_str_dict)
    httpsOnly = fields.Bool(missing=True)
    mountpoints = fields.List(fields.Nested(Mountpoint), validate=validate_mountpoints, missing=[])
    ports = fields.List(fields.Nested(ServicePort), required=True)

    replicas = fields.Int(missing=1)
    minReadySeconds = fields.Int()
    progressDeadlineSeconds = fields.Int()
    strategy = fields.Nested(UpdateStrategy)
    hostAliases = fields.List(fields.Nested(HostAliases))

    containers = fields.List(fields.Nested(ContainerSpec), required=True)
    volumes = fields.List(fields.Dict(), validate=validate_pod_volumes, missing=[])
    hpa = fields.Nested(HPA)
    @post_load
    def finalize(self, data):
        """add defaults to fields, and then construct a Dict"""
        # validate service port
        container_ports = [p for cont in data["containers"] for p in cont["ports"] ]
        port_int_list, port_str_list = [], []
        for p in container_ports:
            if p.get("name") is not None:
                port_str_list.append(p.get('name'))
            port_int_list.append(p['containerPort'])

        svc_ports = data.get("ports", [])
        for p in svc_ports:
            targetPort = p.get("targetPort")
            if targetPort is None:
                continue
            if isinstance(targetPort, str):
                if targetPort not in port_str_list:
                    raise ValidationError("service targetPort doesn't stay in Container port list")
            elif isinstance(targetPort, int):
                if targetPort not in port_int_list:
                    raise ValidationError("service targetPort doesn't stay in Container port list")
        return Dict(data)


service_schema = ServiceSchema()


class TestEntrypointSchema(StrictSchema):
    image = fields.Str()
    volumes = fields.List(fields.Str(), validate=validate_docker_volumes)
    script = fields.List(fields.Str(), required=True)


class TestSchema(StrictSchema):
    builds = fields.List(fields.Nested(BuildSchema), missing=[])
    entrypoints = fields.List(fields.Nested(TestEntrypointSchema))


class AppSpecsSchema(StrictSchema):
    appname = fields.Str(required=True, validate=validate_appname)
    type = fields.Str(missing="worker", validate=validate_app_type)
    builds = fields.List(fields.Nested(BuildSchema), missing=[])
    service = fields.Nested(ServiceSchema, required=True)
    test = fields.Nested(TestSchema)

    @post_load
    def finalize(self, data):
        """add defaults to fields, and then construct a Dict"""
        build_names = set()
        for build in data["builds"]:
            name = build.get("name", None)
            if name:
                if name in build_names:
                    raise ValidationError("duplicate build name")
                build_names.add(name)

        if data["type"] == "web":
            ports = data["service"]["ports"]
            if len(ports) != 1:
                ValidationError("web service should contain only one port")
            for p in ports:
                if p["port"] != 80:
                    raise ValidationError("port of web service must be 80")
        return Dict(data)

    @validates_schema
    def validate_misc(self, data):
        # check raw related fields
        pass


app_specs_schema = AppSpecsSchema()


class JobSchema(StrictSchema):
    jobname = fields.Str(required=True, validate=validate_jobname)
    # the below 4 fields are not used by kubernetes
    git = fields.Str()
    branch = fields.Str(missing='master')
    commit = fields.Str()
    comment = fields.Str()

    backoffLimit = fields.Int()
    completions = fields.Int()

    parallelism = fields.Int()
    autoRestart = fields.Bool(missing=False)

    containers = fields.List(fields.Nested(ContainerSpec), required=True)

    @post_load
    def finalize(self, data):
        """add defaults to fields, and then construct a Box"""
        return Dict(data)


job_schema = JobSchema()


def load_job_specs(raw_data):
    """
    add defaults to fields, and then construct a Dict
    :param raw_data:
    :param tag: release tag
    :return:
    """
    data = job_schema.load(raw_data).data
    return Dict(data)


class SparkAppDriverSchema(StrictSchema):
    cpu = fields.Int(missing=1)
    memory = fields.Str(missing='512m')


class SparkAppExecutorSchema(StrictSchema):
    cpu = fields.Int(missing=1)
    memory = fields.Str(missing='512m')
    instances = fields.Int(missing=1)


class DependencesSchema(StrictSchema):
    jars = fields.List(fields.Str())
    files = fields.List(fields.Str())
    pyFiles = fields.List(fields.Str())


class SparkAppSchema(StrictSchema):
    apptype = fields.Str(required=True, validate=validate_spark_apptype)
    appname = fields.Str(required=True)
    role = fields.Str()
    schedule = fields.Str(validate=validate_schedule)
    concurrencyPolicy = fields.Str(missing='Allow')
    type = fields.Str(validate=validate_spark_type, missing='Python')
    image = fields.Str(required=True)
    imagePullPolicy = fields.Str(missing='Always')
    sparkVersion = fields.Str(missing='2.4.0')
    arguments = fields.List(fields.Str())
    nodeSelector = fields.Dict()
    pythonVersion = fields.Str(missing='3')
    driver = fields.Nested(SparkAppDriverSchema)
    executor = fields.Nested(SparkAppExecutorSchema)
    mainApplicationFile = fields.Str(required=True)
    deps = fields.Nested(DependencesSchema)
    mode = fields.Str(missing='cluster', validate=validate_spark_mode)
    hadoopConfigMap = fields.Str()
    serviceAccount = fields.Str()
    sparkConfigMap = fields.Str()
    comment = fields.Str()
    sparkConf = fields.Dict()
    hadoopConf = fields.Dict()

    @post_load
    def finalize(self, data):
        if data["apptype"] == "scheduledsparkapplication":
            if "schedule" not in data.keys():
                raise ValidationError("scheduledspark app must have `schedule` field")
        return Dict(data)


sparkapp_schema = SparkAppSchema()


def load_sparkapp_specs(raw_data):
    data = sparkapp_schema.load(raw_data).data
    return Dict(data)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

import sys
import re
import os
from os.path import dirname, abspath
from setuptools import setup, find_packages


def find_version(*paths):
    fname = os.path.join(*paths)
    with open(fname) as fhandler:
        version_file = fhandler.read()
        version_match = re.search(r"^__VERSION__ = ['\"]([^'\"]*)['\"]",
                                  version_file, re.M)

    if not version_match:
        raise RuntimeError("Unable to find version string in %s" % (fname,))

    version = version_match.group(1)

    return version


def find_readme(*paths):
    with open(os.path.join(*paths)) as f:
        return f.read()


version = find_version('kaelib', '__init__.py')

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []

requires = [
    'requests>=2.20.0',
    'websocket-client==0.54.0',
    'humanfriendly==4.17',
    'addict==2.2.0',
    'marshmallow==2.17.0',
]
root_dir = dirname(abspath(__file__))

setup(
    name='kaelib',
    version=version,
    description='kae api library',
    long_description=find_readme('README.md'),
    long_description_content_type='text/markdown',
    author='Yu Yang',
    author_email='yangyu@geetest.com',
    url='https://github.com/kaecloud/kaelib',
    include_package_data=True,
    packages=find_packages(root_dir),
    install_requires=requires,
    setup_requires=pytest_runner,
    tests_require=[
        "pyyaml",
        "pytest-cov",
        "pytest-randomly",
        "pytest-mock",
        "pytest>3.0",
    ],
)
